# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from typing import List

import numpy as np


def check_rule(anyon1: int, anyon2: int, outcome: int) -> bool:
    """ Returns True if 'anyon1 x anyon2 = outcome' obeys the Fibonacci
    fusion rules, returns False otherwise.

    Parameters
    ----------
    anyon1 : int
        Anyon charge of the 1st anyon.
    anyon2 : int
        Anyon charge of the 2nd anyon.
    outcome : int
        Anyon charge of the fusion result.

    Returns
    -------
    bool
        True if the Fibonacci fusion rules are obeyed, False otherwise.

    """
    if anyon1 and anyon2:
        return True
    elif (anyon1 or anyon2) and outcome == 1:
        return True
    elif not (anyon1 or anyon2) and outcome == 0:
        return True
    else:
        return False


def check_outcomes(outcomes: List[int]) -> bool:
    previous_outcome = 1

    for outcome in outcomes:
        if check_rule(previous_outcome, 1, outcome):
            previous_outcome = outcome
        else:
            return False
    return True


def check_state(state) -> bool:
    nb_qudits = len(state["qudits"])
    qudit_len = len(state["qudits"][0])

    for qudit in state["qudits"]:
        if len(qudit) == qudit_len:
            if not check_outcomes(qudit):
                return False
        else:
            return False

    if nb_qudits != len(state["roots"]) + 1:
        return False

    previous_outcome = state["qudits"][0][-1]

    for i, outcome in enumerate(state["roots"]):
        if check_rule(previous_outcome, state["qudits"][i + 1][-1], outcome):
            previous_outcome = outcome
        else:
            return False
    return True


def gen_state(comb: List[int], nb_qudits: int, qudit_len: int):
    state = {"qudits": [], "roots": []}

    for i, label in enumerate(comb):
        if i < nb_qudits * qudit_len:
            if i % qudit_len:
                state["qudits"][-1].append(label)
            else:
                state["qudits"].append([label])
        else:
            state["roots"].append(label)

    return state


def generate_basis(nb_qudits: int, nb_anyons_per_qudit: int):
    """Generates all the basis states for a system of a given number of
    qudits, and a given number of anyons per qudit.

    Parameters
    ----------
    nb_qudits : int
        Number of qudits in the circuit.
    nb_anyons_per_qudit : int
        Number of anyons in each qudit.

    Returns
    -------
    basis : List[basis states]
        A list of basis states.
    """
    nb_roots = nb_qudits - 1
    qudit_len = nb_anyons_per_qudit - 1
    nb_labels = nb_qudits * qudit_len + nb_roots

    basis = []

    curr_comb = [0] * nb_labels
    final_comb = [1] * nb_labels

    curr_state = gen_state(curr_comb, nb_qudits, qudit_len)

    if check_state(curr_state):
        basis.append(curr_state)

    while not np.all(curr_comb == final_comb):
        for i, label in enumerate(curr_comb):
            if label == 0:
                curr_comb[i] = 1
                break
            else:
                curr_comb[i] = 0

        curr_state = gen_state(curr_comb, nb_qudits, qudit_len)

        if check_state(curr_state):
            basis.append(curr_state)

    return basis



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np

from .drawer_anyon import DrawerAnyon
from .utils import matplotlib_close_if_inline


class Drawer:
    def __init__(
        self,
        nb_qudits: int,
        nb_anyons_per_qudit: int,
    ):
        self._i = np.linspace(0, 1)
        self._one = np.ones_like(self._i)
        self._hi = np.linspace(0, 0.46, 25)
        self._id = 2 * (0.5 - self._hi[-1])

        self.__nb_qudits = nb_qudits
        self.__nb_anyons_per_qudit = nb_anyons_per_qudit
        self.__nb_anyons = nb_qudits * nb_anyons_per_qudit

        self.__anyons = {}
        self.__idx_map = {}

        self.__STARTING_INDEX = 1

        for i in range(self.__nb_qudits):
            for j in range(self.__nb_anyons_per_qudit):
                id = i * self.__nb_anyons_per_qudit + j + self.__STARTING_INDEX
                self.__idx_map[id] = id
                self.__anyons[id] = DrawerAnyon(
                    id, (i * (self.__nb_anyons_per_qudit + 1) + j)
                )

    @property
    def anyons(self):
        return self.__anyons

    def _sigmoid(self, x):
        expr = 1 / (1 + np.exp(-(x - 0.5) * 12))
        expr -= expr[0]
        expr /= expr[-1]
        return expr

    def braid(self, m: int, n: int) -> None:

        # Preprocess
        m_init = self.__idx_map[m]
        n_init = self.__idx_map[n]

        anyon_over = self.__anyons[m_init]
        anyon_under = self.__anyons[n_init]
        anyons_idle = map(
            lambda x: self.__anyons[x],
            filter(lambda x: x != m_init and x != n_init, self.__anyons.keys()),
        )

        distance = abs(anyon_over.get_last_y() - anyon_under.get_last_y())

        # Process the idle anyons
        for curr_anyon in anyons_idle:
            curr_anyon.add_identity()

        # Process the over anyon
        anyon_over.x = anyon_over.get_last_x() + self._i
        anyon_over.y = anyon_over.get_last_y() - distance * np.sign(
            m - n
        ) * self._sigmoid(self._i)

        # Process the bottom anyon
        anyon_under.x = anyon_under.get_last_x() + self._hi
        anyon_under.x = anyon_under.get_last_x() + self._id + self._hi

        sigm = self._sigmoid(np.append(self._hi, np.linspace(1 - self._hi[-1], 1, 25)))

        start_y = anyon_under.get_last_y()

        anyon_under.y = start_y + distance * np.sign(m - n) * sigm[:25]

        anyon_under.y = start_y + distance * np.sign(m - n) * sigm[25:]

        # Renaming
        self.__idx_map[n], self.__idx_map[m] = self.__idx_map[m], self.__idx_map[n]

    def __fuse(self, idx_anyon_top, idx_anyon_bot):
        self.__anyons[idx_anyon_bot].x = (
            self.__anyons[idx_anyon_bot].get_last_x() + self._i
        )

        self.__anyons[idx_anyon_top].x = (
            self.__anyons[idx_anyon_top].get_last_x() + self._i
        )

        distance = 0.5 * abs(
            self.__anyons[idx_anyon_bot].get_last_y()
            - self.__anyons[idx_anyon_top].get_last_y()
        )

        self.__anyons[idx_anyon_bot].y = self.__anyons[
            idx_anyon_bot
        ].get_last_y() + distance * self._sigmoid(self._i)

        self.__anyons[idx_anyon_top].y = self.__anyons[
            idx_anyon_top
        ].get_last_y() - distance * self._sigmoid(self._i)

    def measure(self):
        for curr_anyon in self.__anyons.values():
            curr_anyon.add_identity()

        # Fusing anyons within qudits
        for i in range(self.__nb_qudits):
            for j in range(self.__nb_anyons_per_qudit - 1):
                # Idle anyons
                for k in range(j + 2, self.__nb_anyons_per_qudit):
                    final_idx = (
                        i * self.__nb_anyons_per_qudit + k + self.__STARTING_INDEX
                    )
                    idx = self.__idx_map[final_idx]
                    self.__anyons[idx].add_identity()

                # Fusing
                final_bot_idx = (
                    i * self.__nb_anyons_per_qudit + j + self.__STARTING_INDEX
                )

                idx_anyon_bot = self.__idx_map[final_bot_idx]
                idx_anyon_top = self.__idx_map[final_bot_idx + 1]

                self.__fuse(idx_anyon_top, idx_anyon_bot)

        # Fusing qudits
        for i in range(1, self.__nb_qudits):
            # 1 -> None
            # 2 -> 1
            # 3 -> 1, 2

            # Idle qudits
            for k in range(i + 1, self.__nb_qudits):
                # 2 -> None
                # 3 -> 2
                final_idx = (
                    (k + 1) * self.__nb_anyons_per_qudit + self.__STARTING_INDEX - 1
                )
                idx = self.__idx_map[final_idx]
                self.__anyons[idx].add_identity()

            # Fusing
            final_bot_idx = i * self.__nb_anyons_per_qudit + self.__STARTING_INDEX - 1
            final_top_idx = (
                (i + 1) * self.__nb_anyons_per_qudit + self.__STARTING_INDEX - 1
            )

            idx_anyon_bot = self.__idx_map[final_bot_idx]
            idx_anyon_top = self.__idx_map[final_top_idx]

            self.__fuse(idx_anyon_top, idx_anyon_bot)

    def draw(self):
        width = self.__anyons[1].get_last_x() * 0.5
        height = self.__nb_anyons * 0.3
        fig, ax = plt.subplots(1, 1, figsize=(width, height))

        for i in range(self.__nb_qudits):
            for j in range(self.__nb_anyons_per_qudit):
                k = i * self.__nb_anyons_per_qudit + j + 1
                curr_anyon = self.__anyons[k]

                for x, y in zip(curr_anyon.x, curr_anyon.y):
                    ax.plot(x, y, curr_anyon.color)
                ax.text(
                    -0.2,
                    curr_anyon.get_first_y(),
                    curr_anyon.label,
                    horizontalalignment="right",
                )
        ax.axis("off")
        if fig:
            matplotlib_close_if_inline(fig)
        return fig



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os

PROGRAM_NAME = "tqsim"
CONFIG_PATH = os.path.join(os.path.expanduser("~"), f".{PROGRAM_NAME}")
STORE_PATH = os.path.join(CONFIG_PATH, "store")



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import matplotlib.pyplot as plt
import matplotlib as mpl
from math import pi, cos, exp, atan


MATPLOTLIB_INLINE_BACKENDS = {
    "module://ipykernel.pylab.backend_inline",
    "module://matplotlib_inline.backend_inline",
    "nbAgg",
}


def matplotlib_close_if_inline(figure):
    """Close the given matplotlib figure if the backend in use draws figures inline.
    If the backend does not draw figures inline, this does nothing.  This function is to prevent
    duplicate images appearing; the inline backends will capture the figure in preparation and
    display it as well, whereas the drawers want to return the figure to be displayed."""
    # This can only called if figure has already been created, so matplotlib must exist.
    import matplotlib.pyplot

    if matplotlib.get_backend() in MATPLOTLIB_INLINE_BACKENDS:
        matplotlib.pyplot.close(figure)


def cplot(cmatrix, sigma=0.5, title=''):
    """Plots a complex-valued matrix with color coding, and a color map.
    'Sigma' controls how much small values are colored. A lower value will
    emphasize small values more.

    Parameters
    ----------
    cmatrix : ndarray
        A complex-valued matrix.
    sigma : float, optional
        Standard deviation squared. The default is 0.5.
    title : str, optional
        Title of the plotted figure. The default is ''.

    Returns
    -------
    None.

    """
    img = []
    for r, row in enumerate(cmatrix):
        img.append([])
        for c in row:
            y = c.imag
            x = c.real

            if x == 0:
                if y > 0:
                    theta = pi/2
                else:
                    theta = -pi/2
            else:
                theta = atan(y/x)
            if x < 0:
                theta += pi
            rad = 1- exp(-(x**2 + y**2)/sigma)
            img[-1].append([cos(theta/2)**2,
                            cos(theta/2 + 2*pi/3)**2,
                            cos(theta/2 - 2*pi/3)**2,
                            rad])

    mpl.rcParams['figure.figsize'] = (10, 10)
    fig, (pl, sc) = plt.subplots(nrows=1, ncols=2, sharex=False)
                                 #figsize=[8, 25])
    sc.imshow(img)
    pl.imshow(scale(sigma=sigma), extent=(-1, 1, -1, 1))
    pl.set_xlabel('Re')
    pl.set_ylabel('Img')
    pl.grid(True)
    plt.title(title)
    plt.show()
    return

def scale(sigma=0.5):
    """
    Plot the scaling spectrum of the complex plane [-1, 1, -i, i]
    """
    img = []
    sc = 50
    for r in range(sc, -sc, -1):
        img.append([])
        for c  in range(-sc, sc, 1):
            y = (r/sc)
            x = (c/sc)
            if x == 0:
                if y > 0:
                    theta = pi/2
                else:
                    theta = -pi/2
            else:
                theta = atan(y/x)
            if x < 0:
                theta += pi
            rad = 1- exp(-(x**2 + y**2)/sigma)
            img[-1].append([cos(theta/2)**2,
                            cos(theta/2 + 2*pi/3)**2,
                            cos(theta/2 - 2*pi/3)**2,
                            rad])

    return img



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import pickle
from typing import List, Sequence, Tuple

import numpy as np

from .config import STORE_PATH  # For caching the bases and sigmas.
from .lib.basis_generator import generate_basis
from .lib.drawer import Drawer
from .lib.operator_generator import generate_braiding_operator


class AnyonicCircuit:
    """This class represents an anyon-based topological quantum circuit.
    Such a circuit is described by the number of qudits it contains, and
    the number of anyons each qudit contains.

    Example: >>>circuit = AnyonicCircuit(nb_qudits=2, nb_anyons_per_qudit=3)
                This circuit has 2 qudits, with 3 anyons each. The circuit has
                a total of 2*3=6 anyons.

    Parameters
    ----------
    nb_qudits : int, optional
        Number of qudits in the circuit. The default is 1.
    nb_anyons_per_qudit : int, optional
        Number of anyons in each qudit. The default is 3.

    Attributes
    ----------
    nb_qudits : int
        Number of qudits in the circuit.
    nb_anyons_per_qudit : int
        Number of anyons in each qudit.
    drawer : Drawer
        A drawer object that handles the drawing of the quantum circuit.
    dim : int
        The dimension of the fusion space for the quantum circuit.
    basis : List
        List of all the basis states.

    """

    def __init__(self, nb_qudits: int = 1, nb_anyons_per_qudit: int = 3):
        """
        Parameters
        ----------
        nb_qudits : int, optional
            Number of qudits in the circuit. The default is 1.
        nb_anyons_per_qudit : int, optional
            Number of anyons in each qudit. The default is 3.

        Returns
        -------
        None.

        """
        self.__nb_qudits = nb_qudits
        self.__nb_anyons_per_qudit = nb_anyons_per_qudit
        self.__nb_anyons = nb_qudits * nb_anyons_per_qudit

        self.__nb_braids: int = 0
        self.__braids_history: List[Tuple[int, int]] = []
        self.__measured: bool = False

        self.__basis, self.__dim = self.__get_basis()

        input_state = np.zeros((self.__dim, 1), dtype=np.complex128)
        input_state[0, 0] = 1
        self.__initial_state = input_state

        self.__sigmas = self.__get_sigmas()
        self.__unitary = np.eye(self.__dim)

        self.__drawer = Drawer(nb_qudits, nb_anyons_per_qudit)

    @property
    def nb_qudits(self):
        """Returns the number of qudits in the circuit.

        Returns
        -------
        int
            Number of qudits.
        """
        return self.__nb_qudits

    @property
    def nb_anyons_per_qudits(self):
        """Returns the number of anyons for each qudit in the circuit.

        Returns
        -------
        int
            Number of anyons per qudit.
        """
        return self.__nb_anyons_per_qudit

    @property
    def drawer(self):
        """Returns the drawer object for the circuit.

        Returns
        -------
        Drawer
            The circuit's drawer object.
        """
        return self.__drawer

    @property
    def dim(self):
        """Returns the dimension of the fusion space.

        Returns
        -------
        int
            Dimension of the fusion space.
        """
        return self.__dim

    @property
    def basis(self):
        """Returns a list of all the basis states for the circuit.

        Returns
        -------
        List
            List of all the basis states.

        """
        return self.__basis

    @property
    def braiding_operators(self):
        """Returns a list of all the braiding operators.

        Returns
        -------
        List
            List of all the braiding operators.

        """
        return self.__sigmas

    def __get_basis(self) -> Tuple[np.ndarray, int]:
        folder_path = os.path.join(
            STORE_PATH, f"{self.__nb_qudits}_{self.__nb_anyons_per_qudit}"
        )
        filename = os.path.join(folder_path, "basis.dat")
        try:
            with open(filename, "rb") as f:
                basis = pickle.load(f)
        except FileNotFoundError:
            basis = generate_basis(self.__nb_qudits, self.__nb_anyons_per_qudit)
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, "wb") as f:
                pickle.dump(basis, f)

        return basis, len(basis)

    def __get_sigmas(self) -> List[np.ndarray]:

        folder_path = os.path.join(
            STORE_PATH, f"{self.__nb_qudits}_{self.__nb_anyons_per_qudit}"
        )
        filename = os.path.join(folder_path, "sigmas.dat")
        try:
            with open(filename, "rb") as f:
                sigmas = pickle.load(f)
        except FileNotFoundError:
            sigmas = []
            for index in range(1, self.__nb_anyons):
                sigma = generate_braiding_operator(
                    index, self.__nb_qudits, self.__nb_anyons_per_qudit
                )
                sigmas.append(np.array(sigma))

            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, "wb") as f:
                pickle.dump(sigmas, f)

        return sigmas

    def initialize(self, input_state: np.ndarray):
        """Initializes the circuit in the state input_state.

        Parameters
        ----------
        input_state : np.ndarray
            A normalized quantum state with the same dimensions as the
            fusion space.
            Example:    for a 1-qudit circuit with 3 anyons, input_state must
                        be a 3 dimensional vector with norm 1.

        Raises
        ------
        Exception
            Will be raised if an initialization is attempted after performing
            braiding operations.
        ValueError
            Will be raised if the input state has the wrong dimension or
            is not normalized.

        Returns
        -------
        AnyonicCircuit
            A reference to the same circuit.
        """
        if self.__nb_braids > 0:
            raise Exception(
                "Initialization should happen before any braiding operation is performed!"
            )

        input_state = np.array(input_state)
        if not np.size(input_state) == self.__dim:
            raise ValueError(f"The state has wrong dimension. Should be {self.__dim}")

        norm = np.sum(np.real(input_state * input_state.conjugate()))
        if not np.isclose(norm, 1):
            raise ValueError("The input state is not normalized correctly!")

        self.__initial_state = np.reshape(input_state, (self.__dim, 1))

        return self

    def braid(self, n: int, m: int):
        """Braids the two anyons at positions 'n' and 'm'.
        If n < m, they are braided in a clockwise direction,
        if n > m, they are braided in a counterclockwise direction.

        Parameters
        ----------
        n : int
            The 1st anyon's position.
        m : int
            The 2nd anyon's position.

        Raises
        ------
        Exception
            Exceptions are raised if a braiding is attempted after a
            measurement, or if trying to braid two non-adjacent anyons.
        ValueError
            Is raised if the parameters are not strictly positive integers,
            or if incorrect positions are passed.

        Returns
        -------
        AnyonicCircuit
            A reference to the same circuit.

        """
        if self.__measured:
            raise Exception("System already measured! Cannot perform further braiding!")

        if not isinstance(m, int) or not isinstance(n, int):
            raise ValueError("n, m must be integers")

        if m < 1 or n < 1:
            raise ValueError("n, m must be higher than 0!")

        if m > self.__nb_anyons or n > self.__nb_anyons:
            ## Check if correct
            raise ValueError(
                f"The system has only {self.__nb_anyons} anyons! n, m are erroneous!"
            )

        if abs(n - m) != 1:
            raise Exception("You can only braid adjacent anyons!")

        if n < m:
            self.__unitary = self.__sigmas[n - 1] @ self.__unitary
        else:
            self.__unitary = self.__sigmas[m - 1].T.conjugate() @ self.__unitary

        self.__braids_history.append((n, m))

        self.__nb_braids += 1

        self.drawer.braid(m, n)

        return self

    def braid_sequence(self, braid: Sequence[Sequence[int]]):
        """Takes a sequence of [sigma operator, power], and applies the
        successive operators to the 'power'.
        The first operator in the sequence is the first to be applied.

        Parameters
        ----------
        braid : Sequence[Sequence[int]]
            A sequence of pairs of integers representing a braiding operator
            and an exponent.

        Raises
        ------
        ValueError
            Is raised if one of the operators' indices is not an integer
            greater or equal to 1, or is an incorrect index.

        Returns
        -------
        AnyonicCircuit
            A reference to the same circuit.
        """
        for ind, power in braid:
            if not isinstance(ind, int) or not isinstance(power, int):
                raise ValueError("Indices and powers must be integers!")
            if ind < 1:
                raise ValueError(
                    f"sigma_{ind} is not a valid braiding operator! "
                    f"The operators indices must be >= 1."
                )
            if ind >= self.__nb_anyons:
                raise ValueError(
                    f"sigma_{ind} is not a valid braiding operator! "
                    f"The operators indices must be < {self.__nb_anyons}."
                )
            # Computing m and n
            m = n = 0
            if power > 0:
                n = ind
                m = ind + 1
            elif power < 0:
                m = ind
                n = ind + 1
            else:  # if power=0, do nothing (identity)
                continue

            for _ in range(abs(power)):
                self.braid(n, m)

        return self

    def measure(self):
        """Performs a measurement on the whole circuit.

        Raises
        ------
        Exception
            Is raised if a measurement has already been carried.

        Returns
        -------
        AnyonicCircuit
            A reference to the same circuit.
        """
        if self.__measured:
            raise Exception("Cannot carry the measurements twice!")

        self.__measured = True
        self.drawer.measure()
        return self

    def history(self, output: str = "raw"):
        """Returns the history of all braiding operations that were performed
        in the circuit.
        Its output can either be the raw braiding operations (n, m), a list of
        braiding operators (sigmas), or a LaTeX string containing the product
        of all the braiding operators.

        Parameters
        ----------
        output : str, optional
            Can either be "raw", "sigmas", or "latex". The default is "raw".

        Raises
        ------
        ValueError
            Is raised if an incorrect output format is chosen.

        Returns
        -------
        List or String
            Either a list of (n, m) operations, a list of braiding operators,
            or a LaTeX string.

        """
        if not output in ["raw", "sigmas", "latex"]:
            raise ValueError('Output should be either: "raw", "sigmas" or "latex"')

        if output == "raw":
            return self.__braids_history

        elif output == "sigmas":
            ret = []
            for (n, m) in self.__braids_history:
                if m < n:
                    ret.append(f"is{m}")
                else:
                    ret.append(f"s{n}")
            return ret

        else:
            sigmas = self.history(output="sigmas")

            # Converting to a sigma notation with powers
            power_sigmas = []
            last_sigma = sigmas[-1] if len(sigmas) else None
            power = 0
            for sigma in reversed(sigmas):
                if sigma == last_sigma:
                    power += 1
                    continue
                # Done counting the powers of the last sigma, adding it
                power_sigmas.append((last_sigma, power))
                # resetting
                power = 1
                last_sigma = sigma

            # Handling the last sigma
            if power != 0:
                power_sigmas.append((last_sigma, power))

            # Converting to LaTeX
            latex = ""
            for sigma, p in power_sigmas:
                # Inverses of sigmas (negative powers)
                if sigma[0] == "i":
                    latex += r"\sigma_{" f"{sigma[2:]}" "}^{" f"{-p}" "}"
                # Sigmas (positive powers)
                else:
                    latex += r"\sigma_{" f"{sigma[1:]}" "}"
                    if p > 1:  # Only add exponents != 1
                        latex += "^{" f"{p}" "}"

            latex = "$ " + latex + "$"
            return latex

    def draw(self):
        """Draws the topological quantum circuit.

        Returns
        -------
        Figure
            The braid describing the topological quantum circuit.
        """
        return self.drawer.draw()

    def statevector(self) -> np.ndarray:
        """Computes and returns the current state vector of the circuit.

        Returns
        -------
        ndarray
            The state vector of the circuit.

        """
        return self.__unitary @ self.__initial_state

    def unitary(self) -> np.ndarray:
        """Returns the unitary representation of the quantum circuit.

        Returns
        -------
        ndarray
            The unitary matrix for the circuit.
        """
        return self.__unitary

    def run(self, shots: int = 1000):
        """Simulates the quantum circuit for 'shots' number of times and
        returns the measurement results.

        Parameters
        ----------
        shots : int, optional
            Number of times the circuit is simulated. The default is 1000.

        Raises
        ------
        Exception
            Is raised if the circuit is run without a measurement.

        Returns
        -------
        dict
            Contains the number of measurements for each measured state.

        """
        # Needs to be measured?
        if not self.__measured:
            raise Exception("The system was not measured!")

        statevector = self.statevector()
        probs = np.ravel(np.real(statevector * statevector.conjugate()))
        memory = np.random.choice(np.arange(self.__dim), p=probs, size=shots)

        idx, counts = np.unique(memory, return_counts=True)

        counts_dict = {}
        for i in range(len(idx)):
            counts_dict[str(idx[i])] = counts[i]

        return {"counts": counts_dict, "memory": memory}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tqsim', 'tqsim.lib']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.6.1,<4.0.0', 'numpy>=1.23.4,<2.0.0']

setup_kwargs = {
    'name': 'tqsim',
    'version': '0.0.2',
    'description': 'TQSim is a Topological Quantum Computing simulator based on Anyon models',
    'long_description': '<img align="right" width="70" src="https://raw.githubusercontent.com/Constantine-Quantum-Tech/tqsim/main/images/cqtech_logo.png" alt="CQTech">\n\n# TQSim\n\n[![CircleCI](https://dl.circleci.com/status-badge/img/gh/Constantine-Quantum-Tech/tqsim/tree/main.svg?style=svg)](https://dl.circleci.com/status-badge/redirect/gh/Constantine-Quantum-Tech/tqsim/tree/main)\n[![Version](https://img.shields.io/pypi/v/tqsim?style=flat-square)](https://pypi.org/project/tqsim/)\n[![License](https://img.shields.io/pypi/dm/tqsim?style=flat-square)](https://pypi.org/project/tqsim/)\n[![License](https://img.shields.io/github/license/Constantine-Quantum-Tech/tqsim?style=flat-square)](LICENSE)\n\n\nTQSim stands for Topological Quantum Simulator. It is an open-source library for simulating topological quantum computer based on anyons. \n\n\n## Installation\n\nYou can install TQSim from pip using\n\n```bash\npip install --upgrade tqsim\n```\n\n## Usage\n\n### 1. Basic Example\nIn this example, we create a circuit with 2 qudits, made of 3 anyons each. We then braid the anyons manually.\n```python\nfrom tqsim import AnyonicCircuit\n\ncircuit = AnyonicCircuit(2, 3) # Create a circuit having 2 qudits and 3 anyons per qudits\ncircuit.braid(1, 2) # Braids the first with the second anyon\ncircuit.braid(3, 4) # Braids the first with the second anyon\ncircuit.braid(2, 1)\ncircuit.measure() # Measure the system by fusing the anyons\ncircuit.draw() # Draw the circuit\n```\nHere is the output of the `draw()` method:\n\n![Circuit Output](https://i.ibb.co/3z9pFmQ/example.png)\n\nSimulating the circuit:\n\n```python\ncircuit.run(shots = 50)\n\n```\n\nOutput:\n```bash\n{\'counts\': {\'0\': 16, \'2\': 20, \'4\': 14}, \'memory\': array([4, 2, 2, 2, 2, 2, 2, 4, 2, 4, 2, 0, 0, 0, 4, 0, 4, 0, 0, 0, 0, 4,\n       2, 4, 0, 2, 0, 0, 0, 0, 4, 4, 2, 2, 2, 4, 2, 2, 0, 0, 2, 4, 2, 2,\n       4, 2, 4, 4, 0, 2])}\n```\n\n### 2. Simulating a Hadamard gate\nHere we simulate the application of a Hadamard gate on a single qudit with 3 anyons.\nUnlike the previous example, we will use a braiding sequence of braiding operators and their corresponding powers.\n```python\nfrom tqsim import AnyonicCircuit\n\ncircuit = AnyonicCircuit(nb_qudits=1, nb_anyons_per_qudit=3)  # Create a circuit with 1 qudit composed of 3 anyons\ncircuit.initialize([0,0,1])  # We initialize the circuit in the last state (state 2).\n                            # For this circuit, we have 3 basis states: [0, 1, 2].\n\n# The Hadamard gate braiding sequence in terms of braiding operators\nhad_sequence = [[1, 2], [2, 2], [1, -2], [2, -2], [1, 2], [2, 4], [1, -2], [2, 2],\n                [1, 2], [2, -2], [1, 2], [2, -2], [1, 4]]\n\ncircuit.braid_sequence(had_sequence)  # We apply the braiding sequence.\n                                      # This should put our qudit in a superposition of the states 2 and 1.\ncircuit.measure()  # Measure the system by fusing the anyons\ncircuit.draw()  # Draw the circuit\n```\nThe Hadamard braid looks like this\n\n![Circuit Output](https://i.ibb.co/5kDVWgf/example-hadamard.png)\n\nSimulating the circuit:\n\n```python\nresult = circuit.run(shots = 1000)  # Run the circuit 1000 times.\nprint(result[\'counts\'])  # Show the counts only.\n\n```\n\nOutput:\n```bash\n{\'1\': 493, \'2\': 507}\n```\n\n## Authors and Citation\n*Abdellah Tounsi, Mohamed Messaoud Louamri, Nacer eddine Belaloui, and Mohamed Taha Rouabah – Constantine Quantum Technologies.*\n\nIf you have used TQSim in your work, please use the [BibTeX file](citation.bib) to cite the authors.\n\n## License\n\nCopyright © 2022, [Constantine Quantum Technologies](https://cqtech.org). Released under the [Apache License 2.0](LICENSE).\n',
    'author': 'CQTech.org',
    'author_email': 'constantinequantumtech@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Constantine-Quantum-Tech/tqsim',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np


class DrawerAnyon:
    def __init__(self, initial_id, pos):
        self.__i = np.linspace(0, 1)
        self.__one = np.ones_like(self.__i)
        self.__x = [np.copy(self.__i)]
        self.__y = [pos * self.__one]

        self.__color = "black"
        self.__label = ""

    def __repr__(self):
        return f"{self.__label} (color: {self.__color}"

    def __str__(self):
        return f"{self.__label} (color: {self.__color}"

    @property
    def label(self):
        return self.__label

    @label.setter
    def label(self, new):
        self.__label = new

    @property
    def color(self):
        return self.__color

    @color.setter
    def color(self, new):
        self.__color = new

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, new):
        self.__x.append(new)

    @property
    def y(self):
        return self.__y

    @y.setter
    def y(self, new):
        self.__y.append(new)

    def get_last_x(self):
        return self.__x[-1][-1]

    def get_first_y(self):
        return self.__y[0][0]

    def get_last_y(self):
        return self.__y[-1][-1]

    def add_identity(self):
        self.__x.append(self.__i + self.get_last_x())
        self.__y.append(self.__one * self.get_last_y())



# This code is part of TQSim.
#
# (C) Copyright Constantine Quantum Technologies, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from copy import deepcopy

import numpy as np

from .basis_generator import generate_basis


def F(a1, a2, a3, outcome):
    """
    F matrix
    """
    inv_phi = (np.sqrt(5) - 1) / 2  # inverse of golden number
    f_matrix = np.array([[0, 0], [0, 0]])

    # a1 + a2 + a3 + outcome = 4
    if a1 + a2 + a3 + outcome == 4:
        f_matrix = np.array([[inv_phi, np.sqrt(inv_phi)], [np.sqrt(inv_phi), -inv_phi]])
    # a1 + a2 + a3 + outcome = 3
    elif a1 + a2 + a3 + outcome == 3:
        f_matrix = np.array([[0, 0], [0, 1]])
    # a1 + a2 + a3 + outcome = 2
    elif a1 + a2 + a3 + outcome == 2:
        if a1 + a2 == 2:
            f_matrix = np.array([[0, 1], [0, 0]])
        elif a2 + a3 == 2:
            f_matrix = np.array([[0, 0], [1, 0]])
        elif a1 + a3 == 2:
            f_matrix = np.array([[0, 0], [0, 1]])
        elif a3 + outcome == 2:
            f_matrix = np.array([[0, 1], [0, 0]])
        elif a1 + outcome == 2:
            f_matrix = np.array([[0, 0], [1, 0]])
        elif a2 + outcome == 2:
            f_matrix = np.array([[0, 0], [0, 1]])
    # a1 + a2 + a3 + outcome = 1
    # a1 + a2 + a3 + outcome = 0
    elif a1 + a2 + a3 + outcome == 0:
        f_matrix = np.array([[1, 0], [0, 0]])

    return f_matrix


def R(a1, a2):
    """
    R matrix
    """
    if a1 + a2 == 2:
        r_matrix = np.array(
            [[np.exp(-4 * np.pi * 1j / 5), 0], [0, np.exp(3 * np.pi * 1j / 5)]]
        )
    else:
        r_matrix = np.array([[1, 0], [0, 1]])

    return r_matrix


def B(a0, a1, a2, outcome):
    """
    Braiding matrix
    """
    b_matrix = F(a0, a1, a2, outcome) @ R(a1, a2) @ F(a0, a2, a1, outcome).T.conjugate()

    return b_matrix


def sigma(index, state_f, state_i):
    """
    Amplitude of getting state_f by applying the braiding operator
    sigma_{index} on state_i.

    Returns:
        the component (state_f, state_i) of the sigma_{index} matrix
    """
    if index <= 0 or index > len(state_i):
        raise ValueError("index value is not valid!")

    stt_f = [1] + state_f
    stt_i = [1] + state_i

    if index - 2 < 0:
        a0 = 0
    elif index - 2 == 0:
        a0 = 1
    else:
        a0 = state_i[index - 3]

    outcome = state_i[index - 1]
    a = stt_i[index - 1]
    b = stt_f[index - 1]
    amplitude = B(a0, 1, 1, outcome)[a, b]

    ket = stt_i
    ket[index - 1] = b
    bra = stt_f
    if ket == bra:
        braket = 1
    else:
        braket = 0
    return amplitude * braket


def L(k, h, i_, i, jj_, jj):
    """
    L matrix component that is used in calculation of braiding between
    two anyons separated in two qudits.
    (see report)

    Inputs:
        k: int: k
        h: int: i_{m(q-1)}
        i_: int: i'_{mq}
        i: int: i_{mq}
        jj_: list: [i'_{(m+1)1},....i'_{(m+1)q}]
        jj: list: [i_{(m+1)1},....i_{(m+1)q}]
    """
    component = 0 + 0j

    qudit_len = len(jj)
    jjj_ = deepcopy(jj_)
    jjj = deepcopy(jj)
    jjj_ = [1] + jjj_
    jjj = [1] + jjj

    init_p = [0] * qudit_len
    final_p = [1] * qudit_len
    new_p = init_p
    while new_p != final_p:
        pp = deepcopy(new_p)
        pp.append(k)
        product = 1 + 0j
        for ii in range(qudit_len):
            product = (
                product
                * F(i, jjj[ii], 1, pp[ii + 1]).T.conjugate()[jjj[ii + 1], pp[ii]]
                * F(i_, jjj_[ii], 1, pp[ii + 1])[pp[ii], jjj_[ii + 1]]
            )

        product = product * B(h, 1, 1, pp[0])[i, i_]
        component += product
        # iterate
        for ii, label in enumerate(new_p):
            if label == 0:
                new_p[ii] = 1
                break
            else:
                new_p[ii] = 0

    # final iteration
    pp = deepcopy(new_p)
    pp.append(k)
    product = 1 + 0j
    for ii in range(qudit_len):
        product = (
            product
            * F(i, jjj[ii], 1, pp[ii + 1]).T.conjugate()[jjj[ii + 1], pp[ii]]
            * F(i_, jjj_[ii], 1, pp[ii + 1])[pp[ii], jjj_[ii + 1]]
        )

    product = product * B(h, 1, 1, pp[0])[i, i_]
    component += product

    return component


def S(jm, jmo, jmoo, jmo_, h, i_, i, jj_, jj):
    """
    S matrix or sewing matrix is used in calculation of braiding operator
    between two anyons separated between two qudits not fused imedialtely.

    Inputs:
        jm: int: j_m
        jmo: int: j_{m-1}
        jmoo: int: j_{m-2}
        jmo_: int: j'_{m-1}
        h: int: i_{m(q-1)}
        i_: int: i'_{mq}
        i: int: i_{mq}
        jj_: list: [i'_{(m+1)1},....i'_{(m+1)q}]
        jj: list: [i_{(m+1)1},....i_{(m+1)q}]
    """
    component = 0 + 0j

    for kk in [0, 1]:
        component += (
            F(jmoo, i, jj[-1], jm)[jmo, kk]
            * L(kk, h, i_, i, jj_, jj)
            * F(jmoo, i_, jj_[-1], jm).T.conjugate()[kk, jmo_]
        )

    return component


def gen_sigma(index, state_i, state_f):
    qudit_len = len(state_i["qudits"][0])
    nb_anyons_per_qudit = qudit_len + 1

    amplitude = 0
    braket = 1

    if index % nb_anyons_per_qudit > 0:

        m = index // nb_anyons_per_qudit
        idx = index % nb_anyons_per_qudit
        amplitude = sigma(idx, state_f["qudits"][m], state_i["qudits"][m])

        for i, qudit in enumerate(state_i["qudits"]):
            if i == m:
                continue
            elif qudit != state_f["qudits"][i]:
                braket = 0

        for i, root in enumerate(state_i["roots"]):
            if root != state_f["roots"][i]:
                braket = 0

    else:
        m = (index // nb_anyons_per_qudit) - 1

        new_state_i = deepcopy(state_i)
        new_state_i["qudits"][m][-1] = deepcopy(state_f["qudits"][m][-1])
        new_state_i["qudits"][m + 1] = deepcopy(state_f["qudits"][m + 1])
        """
            jm: int: j_m
            jmo: int: j_{m-1}
            jmoo: int: j_{m-2}
            jmo_: int: j'_{m-1}
            h: int: i_{m(q-1)}
            i_: int: i'_{mq}
            i: int: i_{mq}
            jj_: list: [i'_{(m+1)1},....i'_{(m+1)q}]
            jj: list: [i_{(m+1)1},....i_{(m+1)q}]
        """
        if m + 1 > 2:
            new_state_i["roots"][m - 1] = state_f["roots"][m - 1]

            jj_ = deepcopy(new_state_i["qudits"][m + 1])
            jj = deepcopy(state_i["qudits"][m + 1])
            h = state_i["qudits"][m][-2]
            i = state_i["qudits"][m][-1]
            i_ = new_state_i["qudits"][m][-1]

            jmo_ = new_state_i["roots"][m - 1]
            jmoo = state_i["roots"][m - 2]
            jmo = state_i["roots"][m - 1]
            jm = state_i["roots"][m]

        elif m + 1 == 2:
            new_state_i["roots"][m - 1] = state_f["roots"][m - 1]

            jj_ = deepcopy(new_state_i["qudits"][m + 1])
            jj = deepcopy(state_i["qudits"][m + 1])
            h = state_i["qudits"][m][-2]
            i = state_i["qudits"][m][-1]
            i_ = new_state_i["qudits"][m][-1]

            jmo_ = new_state_i["roots"][m - 1]
            jmoo = state_i["qudits"][0][-1]
            jmo = state_i["roots"][m - 1]
            jm = state_i["roots"][m]

        elif m + 1 == 1:

            jj_ = deepcopy(new_state_i["qudits"][m + 1])
            jj = deepcopy(state_i["qudits"][m + 1])
            h = state_i["qudits"][m][-2]
            i = state_i["qudits"][m][-1]
            i_ = new_state_i["qudits"][m][-1]

            jmo_ = new_state_i["qudits"][0][-1]
            jmoo = 0
            jmo = state_i["qudits"][0][-1]
            jm = state_i["roots"][m]

        amplitude += S(jm, jmo, jmoo, jmo_, h, i_, i, jj_, jj)
        if new_state_i != state_f:
            braket = 0

    return braket * amplitude


def generate_braiding_operator(index: int, nb_qudits: int, nb_anyons_per_qudit: int):
    """Generates the braiding operator of index 'index' for a system of
    a given number of qudits and anyons per qudit.
    This operator braids anyons at positions 'index' and 'index'+1.

    Parameters
    ----------
    index : int
        The operator's index.
    nb_qudits : int
        Number of qudits in the circuit.
    nb_anyons_per_qudit : int
        Number of anyons in each qudit.

    Returns
    -------
    List
        Matrix representation of the braiding operator.

    """
    basis = generate_basis(nb_qudits, nb_anyons_per_qudit)

    sigmas = []
    for f, base_f in enumerate(basis):
        sigmas.append([])
        for base_i in basis:
            sigmas[f].append(gen_sigma(index, base_i, base_f))

    return sigmas
